%% Reliability analysis of power systems
%  Hybrid version 1, based on OPAL prototype (analytical) and State
%  sampling (Monte Carlo Simulation)
%  Written by Trygve V. Berg as part of a Master thesis

%% Initializing test network from Excel file
%  SAMREL 4-area network

tic; %starting clock for runtime

% Location of Excel file for input of case data and output of results
inputfile = 'Case files\4-area_network_extra_line.xls';

% Load MATPOWER case data (network data)
mpc = read_mpc_from_Excel(inputfile);

% Load settings/options for the contingency analysis
contanalysis_settings = read_contanalysis_settings_from_Excel(inputfile);

% Load input data for corrective actions
corrective_actions_data = read_corrective_actions_data_from_Excel(inputfile);

% Setting up MATPOWER options
mpopt = setup_mpopt_opalcons(contanalysis_settings);

% Load input data for operating states
os_data = read_os_data_from_Excel(inputfile);

% Load settings/options for constructing the contingency/outage list
ol_settings = read_ol_settings_from_Excel(inputfile);

% Read delivery point data
[dp_data, ~] = read_loaddata_from_Excel(inputfile);
dp_buses = dp_data.bus;

%% OPAL prototype evaluation of first order contingencies

%Only evaluating the 1st order line outages
ol_settings.branch_outage_order = 1; 

% Construct the contingency/outage list 
ol = create_ol(ol_settings,mpc);

% Run contingency analysis
cont_results = opalcons(mpc, ol, contanalysis_settings, corrective_actions_data, os_data, dp_buses);

% Load general reliability analysis settings from Excel file
reliabilitycalc_settings = read_reliabilitycalc_settings_from_Excel(inputfile);

% Read reliability/outage data
rel_data = read_rel_data_from_Excel(inputfile,reliabilitycalc_settings);

% Read component reliability statistics data
componentdata = read_componentdata(reliabilitycalc_settings.component_database,0,[]);

% Update reliability data with component reliability statistics where relevant
rel_data.outagedata = update_outage_data_from_stats(rel_data.outagedata,componentdata);

% Read delivery point data
[dp_data, ~] = read_loaddata_from_Excel(inputfile);

% Setting up convenient data structure for interruption cost function calculations
customer_database = reliabilitycalc_settings.customer_database;
[intcost_func, dp_data] = create_intcost_func(customer_database,...
    reliabilitycalc_settings,os_data,dp_data);

% Run reliability analysis
year = 1;
[aggres, cutres, cut_idx] = reliabilitycalc(reliabilitycalc_settings,...
    year,cont_results,dp_data,componentdata,rel_data,os_data,intcost_func);


%% State Sampling MCS of higher order contingencies

% Resetting the line outage settings to include higher orders
ol_settings.branch_outage_order = 2; 

% Creating a matrix with line data
%          Line nb   Fbus    Tbus      FOR                   
linedata = [ 1        10001	  10002	   0.07515/8736           ;
             2        10001	  10003	   0.1503/8736            ;
             3        10004	  10002	   0.4509/8736            ;
             4        10003	  10006	   0.1503/8736            ;
             5        10003	  10035	   1.323468/8736          ;
             6        10006	  10004	   0.7515/8736            ;
             7        10007	  10004	   1.323468/8736          ;
             8        10035	  10005	   0.7515/8736            ;
             9        10021   10006    1.323468/8736          ; 
            10        20008   10007    6.9464/8736            ;
            11        30023   10021    3.4732/8736            ;
            12        20009   20008    1.323468/8736          ;
            13        20010   20009    0.1503/8736            ;
            14        20012   20009    0.7515/8736            ;
            15        20011   20010    0.4509/8736            ;
            16        20012   20011    0.3006/8736            ;
            17        20022   20012    1.323468/8736          ;
            18        30013   20022    10.4196/8736           ;
            19        30014   20022    12.1562/8736           ;
            20        30016   30013    1.323468/8736          ;
            21        30015   30014    1.323468/8736          ;
            22        30016   30015    0.7515/8736            ;
            23        30017   30015    0.07515/8736           ;
            24        30018   30016    0.22545/8736           ;
            25        30020   30016    0.07515/8736           ;
            26        30019   30017    0.1503/8736            ;
            27        30019   30018    0.3006/8736            ;
            28        30020   30018    0.07515/8736           ;
            29        30019   40024    0.07515/8736           ;    
            30        30023   30020    1.323468/8736          ];
         
%FOR =(faults per year * repair time )/ hours in a year

% Simulation start

N = 150000 ; %number of simulation years
M = 8736; %hours in a year
L = length(linedata(:,1));%number of lines

lambdadp_a = zeros(length(dp_buses),N); %annual frequency of interruption
durationdp_a = zeros(length(dp_buses),N); %annual duration of interruptions
Pinterrdp_a = zeros(length(dp_buses),N); % annual interrupted power for each dp
EENSdp_a = zeros(length(dp_buses),N);%yearly EENS for each delivery point 

for y = 1:N %loop through all sim years
    i_state = 0; %operating state 
    Ploss = 0; %loss of load due to curtailment
    
    outagecounter = zeros(length(dp_buses),1); %number of outages for each dp for a specific year
    lambdadp_y = zeros(length(dp_buses),1); %specific year frequency of interruption for each dp
    durationdp_y = zeros(length(dp_buses),1); %specific year duration of interruptions for each dp
    Pinterrdp_y = zeros(length(dp_buses),1); %specific year interrupted power for each dp
    EENSdp_y = zeros(length(dp_buses),1);% sum of EENS for each dp for this specific year

    
    
    for i = 1:M %loop through a year in hours
        EENSdp_h = 0; %energy not supplied for the specific contingency in hourly increment
        outagebranches = []; %lines that are on outage
        index = 1; %index used for storing the outage lines in order
        for k = 1:L %loop through all lines
            U = rand(); %random uniform number
            if U < linedata(k,4) % line outage at hourly increment
                outagebranches(index) = k; %adding specific line to the outage vector
                index = index+1;
            end %if U
        end %k
        if length(outagebranches) >= 2
            if i < 2184 % if we are within the first 3 months (in hours) use "heavy load" operating state
                i_state = 1;
            else % we are in any of the other months (in hours) use "light load" operating state
                i_state = 2;
            end %if i <8736
            mpc = init_os(mpc,os_data,i_state); %Initialize the correct os for use in contingency analysis
            Ploss = contanalysis_singlecont_montecarlosim(outagebranches,mpc,dp_buses,mpopt,corrective_actions_data,contanalysis_settings); %consequence analysis of sampled lines
            if sum(Ploss) > 0 %if outage causes power loss
            Pinterrdp_y = Pinterrdp_y + Ploss; % accumulating interrupted power for specific year
            EENSdp_h = Ploss*1; % E = P*deltaT (deltaT = 1 hour for state sampling)
            EENSdp_y = EENSdp_y + EENSdp_h; % adding hourly EENS to total EENS for this year
                for j = 1:length(Ploss) % checking for which dp the duration and frequency happens for
                    if Ploss(j) > 0 %Not an elegant solution, but should work
                        outagecounter(j) = outagecounter(j) + 1; %counting number of outages 
                        durationdp_y(j) = durationdp_y(j) + 1; % accumulating interruption duration for specific year (+1 hour for each contingency)
                    end % if Ploss(j)
                end %for j
            end %if Ploss
        end %if length
    end %i
    for p = 1:length(dp_buses)
        lambdadp_a(p,y) = outagecounter(p)/8736; %calculating and adding number this years interruptions to annual number of intterruptions matrix
        durationdp_a(p,y) = durationdp_y(p); %adding this years outage duration to the annual outage duration matrix
        Pinterrdp_a(p,y) = Pinterrdp_y(p); %adding the interrupted power from the specific year to the annual Pinterr matrix
        EENSdp_a(p,y) = EENSdp_y(p); %adding the EENS from this year to annual EENS matrix
    end %for p
end %y

%% Reliability analysis
%  Combining the results from OPAL and MCS 

% Calculating indices for each dp
Pinterrdp_a_avg = (sum(Pinterrdp_a,2))/N + aggres.sum_dp.Pinterr_a/1000; %dividing the reliabilitycalc part by 1000 because it is in kW (as opposed to MW)
EENSdp_a_avg = (sum(EENSdp_a,2))/N + aggres.sum_dp.ENS_a/1000; 
lambdadp_a_avg = (sum(lambdadp_a,2))/N + aggres.sum_dp.lambda_a;
durationdp_a_avg = (sum(durationdp_a,2))/N + aggres.sum_dp.U_a;

%Calculating system indices
Pinterrsys_a_avg = sum(sum(Pinterrdp_a))/N + aggres.sum_sys.Pinterr_a/1000;
EENSsys_a_avg = sum(sum(EENSdp_a))/N + aggres.sum_sys.ENS_a/1000;
lambdasys_a_avg = sum(sum(lambdadp_a))/N + sum(aggres.sum_dp.lambda_a);
durationsys_a_avg = sum(sum(durationdp_a))/N + sum(aggres.sum_dp.U_a);

% Calculating the standard deviation and CV for the MCS part
EENSsys_tot = sum(sum(EENSdp_a)); %total EENS for all years
EENSsys_avg = sum(EENSsys_tot)/N; % average EENS over N simulation years
EENSsys_a = sum(EENSdp_a); % annual EENS for N sim years
expression = 0;
for o = 1:N
    expression = expression + (EENSsys_a(o)-EENSsys_avg)^2; %summing up parts of standard deviation
end
EENSsys_stdd = sqrt(((1/(N-1))*expression)); %standard deviation
EENSsys_cv = EENSsys_stdd/(sqrt(N)*EENSsys_avg); %coefficient of variation

% Storing all indices in one struct
indices = {}; 
indices.dp = {}; 
indices.sys = {};
[indices.dp.Pinterrdp_a_avg,indices.dp.EENSdp_a_avg,indices.dp.lambdadp_a_avg,indices.dp.durationdp_a_avg] = deal(Pinterrdp_a_avg,EENSdp_a_avg,lambdadp_a_avg,durationdp_a_avg);
[indices.sys.Pinterrsys_a_avg,indices.sys.EENSsys_a_avg,indices.sys.lambdasys_a_avg,indices.sys.durationsys_a_avg,indices.sys.EENSsys_cv] = deal(Pinterrsys_a_avg,EENSsys_a_avg,lambdasys_a_avg,durationsys_a_avg,EENSsys_cv);


toc;
